#!/bin/sh
# \
exec wish "$0" "$@"

# Create a simple text widget. No scrollbars or anything fancy.
#
text .t -setgrid true -wrap word -width 80 -height 25 -bg white

# Create a frame with the previous, mark and next buttons
#
frame .f -relief groove -bd 2
set markLabel "Set mark at 1.0"

button .f.p -text "<<< Previous mark" -command [list skipToMark previous]
button .f.m -textvariable markLabel -command setMark
button .f.n -text "Next mark >>>" -command [list skipToMark next]
pack .f.p .f.m .f.n -side left -expand 1

# Pack the text widget last so it gets priority when the window is resized,
# then ensure the text widget has keyboard focus.
#
pack .f -side bottom -fill x -padx 5 -pady 5 -ipady 5
pack .t -side top -fill both -expand true
focus .t

# Now the tag definitions. markpoint is used to indicate where a mark is. The
# rest are for text formatting.
#
.t tag configure markpoint -background red -relief raised -borderwidth 2

.t tag configure title   -font "helvetica 14" \
                         -justify "center" \
                         -underline on
.t tag configure body    -font "helvetica 10" \
                         -justify "left" \
                         -spacing1 16 -spacing2 8
.t tag configure super   -font "helvetica 6" \
                         -offset 4
.t tag configure signoff -font "helvetica 14 italic" \
                         -justify "right"

# Insert the text. Read this carefully - the tag names are mixed with the
# text! The syntax of this insert command is that the tag, if specified,
# applies to the preceding text in the line.
#
.t insert 1.0 "The Legend of Black Cave\n" title

.t insert end \
"The entrance to Black Cave is a sinister crack in the rock at the foot\
of Deadman's Cliff. It is only accessible at low tide. We know that many\
smugglers drowned in Black Cave, cut off by the incoming sea as they tried\
to hide their ill-gotten gains.
On the night of February 19" body \
"th" super \
", 1876, a notorious smuggler named Jeremiah Booth entered the cave,\
accompanied by 3 of his companions. None of the men who entered the cave\
that night were ever seen again, although it is believed their ghosts\
haunt the cave's entrance in an eternal bid to guard their gold..." body \
"\n...to be continued\n" signoff

# Put the text widget's cursor at the top
#
tk::TextSetCursor .t 1.0

# When an event occurs which has (probably) moved the text widget's cursor
# I update the label on the "set mark" button. I do this when either a key is
# released or a mouse button is released. The button label shows whatever is
# in the "markLabel" variable, so all I need do here is update that.
#
foreach eventType [list KeyRelease ButtonRelease] {
    bind .t <$eventType> {+
	set markLabel "Set mark at [.t index insert]"
    }
}

# This is called when the "set mark" button is clicked. It creates a mark at
# the text cursor point called _mark_n, where 'n' is just an incrementing
# number. To make the mark visible in the text I add the "markpoint" tag
# to the character at the new mark.
#
set markCounter 0
proc setMark {} {

    .t mark set _mark_$::markCounter [.t index insert]
    .t tag add markpoint [.t index insert]
    incr ::markCounter
}

# This is called when either of the "move to mark" buttons are clicked. It 
# takes a direction parameter - either "next" or "previous" - and calls
# the text widget command to get the index of the closest mark in that
# direction. Only marks named "_mark_*" are considered; any others are ignored
# and another mark is looked for.
# Thanks to Bryan Oakley for help with this piece of code.
#
proc skipToMark { direction } {
    set index "insert"
    set startIndex [.t index $index]
    while { $index != "" } {
	set index [.t mark $direction $index]
	if { [string match "_mark_*" $index] &&
	     [.t compare $index != $startIndex] } {
	    tk::TextSetCursor .t $index
	    return
	}
    }
}







#
# It's best you stop reading here... :o)
#

#
# OK then....
#

# This code attempts to correct the silly up/down cursor movement which the
# text widget uses by default. It's a variant of the solution at the Wiki
# page: http://mini.net/tcl/3082
#
# This works by binding the up and down arrow keys of all widgets of the Text
# class to this moveUpDown procedure. Instead of just moving down to the next
# new line, it calculates the location of the character below the insertion
# cursor using the height of the current character. The algorithm isn't
# perfect since it's easily fooled by increased line spacing and sub- and
# super-scripting. It does make things feel a lot more sensible though.
#
# As of this writing, December 16, 2003, this workaround isn't required for
# the Tk in CVS. If Tk-8.5 is out by the time you run this this code, and
# you're running that version, you can safely delete this entire block.
#
proc min args {lindex [lsort -real $args] 0}
proc max args {lindex [lsort -real $args] end}
proc moveUpDown {textWidget upOrDown} {
    # Make the insertion cursor visible so bbox doesn't return empty list
    $textWidget see insert

    # Find the coordinates of the cursor and set the new height
    # manually. Note: errors rounding off, since
    # coordinates don't match character positions exactly.
    scan [$textWidget index insert] {%d.%d} lines char
    scan [$textWidget bbox [$textWidget index insert]] \
	{%d %d %d %d} x y textWidth textHeight
    set spacing 0
    foreach tagName [$textWidget tag names insert] {
	set tagSpacing [$textWidget tag cget $tagName -spacing2]
	if { $tagSpacing != "" } {
	    set spacing [max $spacing $tagSpacing]
	}
    }
    incr textHeight $spacing
    scan [$textWidget bbox \
	      @[winfo width $textWidget],[winfo height $textWidget]] \
	      {%*d %d %*d %*d} maxy
    # When updating position, make sure y is within text boundaries
    switch -- $upOrDown {
	"up" {
	    if {$y <= $textHeight} {
		$textWidget yview scroll -1 units
	    } else {
		set y [max [expr $y-$textHeight] 0]
	    }
	}
	"down" {
	    if {$y >= $maxy} {
                $textWidget yview scroll 1 units
	    } else {
                set y [min [expr $y+$textHeight] $maxy]
	    }
	}
    }
    scan [$textWidget bbox [$textWidget index @$x,$y]] \
	{%d %d %d %*d} newx newy width

    # Test on which side of the character
    # we should position the cursor
    if {$x>[expr $newx+$width/2]} {
	set x [expr $newx+$width+1]
    }
    return [$textWidget index @$x,$y]
}

# Replace the default Text widget bindings to try it out
bind Text <Up> {
    tk::TextSetCursor %W [moveUpDown %W up]
}
bind Text <Down> {
    tk::TextSetCursor %W [moveUpDown %W down]
}

# Selection via the keyboard should be re-bound for consistency
bind Text <Shift-Up> {
    tk::TextKeySelect %W [moveUpDown %W up]
}
bind Text <Shift-Down> {
    tk::TextKeySelect %W [moveUpDown %W down]
}