#!/bin/sh
# \
exec wish "$0" "$@"

# This is the search engine, which is called each time a key is pressed (well,
# released actually) in the search text widget. The cursor is in the main
# text widget at a mark labelled "insert". This is the point the search
# starts from. The cursor is moved to the found text if the text is 
# successful. The insertion cursor mark can be moved back to top of the
# text by hitting F1 in the search text widget (or hitting Home in the main
# text widget of course). F3 in the search text widget searches again from
# the current insertion mark onwards.
#
# The parameter is the X11 key symbol of the key pressed.
#
proc search { keysym } {

    # Check for nothing to find!
    #
    if { [string trim [.f.t get 1.0 end]] == "" } {
	return
    }

    # If they hit F1 in the search box, reset the search back to the start
    # of the text
    #
    if { $keysym == "F1" } {
	tk::TextSetCursor .f.t 1.0
	return
    }

    # If they hit F3 in the search box, move the cursor just past the current
    # insertion mark location
    #
    if { $keysym == "F3" } {
	tk::TextSetCursor .f.t "insert +1 chars"
    }

    # Clear an existing search tag, if there is one
    #
    catch { eval .f.t tag remove search [.f.t tag ranges search]}

    # Fetch the string to look for, less leading or trailing whitespace
    #
    set searchStr [string trim [.sf.e get]]

    # OK, do the search from the current insertion cursor point onwards
    #
    set found [.f.t search -exact -- $searchStr insert end]

    if { $found != "" } {

	# Found! Tag the text from the found point, for however many chars
	# were looked for. Then set the insertion cursor to that point and
	# ensure it's visible in the window.
	#
	.f.t tag add search $found "$found +[string length $searchStr] chars"
	.f.t mark set insert $found
	.f.t see $found
	.sbf.sb configure -text "Found: $searchStr"

    } else {
	.sbf.sb configure -text "Not found: $searchStr"
    }
}

# This is the highlight procedure which highlights the required areas of the
# text. Hightlights are applied/unapplied immediately by changing the
# attributes of the appropriate tags. As long as the tags are in the right
# place in the text, all the colour changes happen instantly.
#
# The changed parameter is the name of the highlight type which has been
# toggled on or off. This will be "open", "close", "empty" etc.
#
proc highlight { changed } {

    # These globals are controlled byt the GUI. Each will be set to 1 or 0
    # depending on whether the onscreen checkbox is checked or cleared.
    #
    global open close empty attributes hrefs

    # This switch looks at the highlight type which has changed, then sets the
    # appropriate tag configuration based on whether the onscreen checkbox is
    # currently set on or off.
    #
    switch $changed {
	open {
	    if { $open } {
		.f.t tag configure openElement -foreground red
	    } else {
		.f.t tag configure openElement -foreground {}
	    }
	}

	close {
	    if { $close } {
		.f.t tag configure closeElement -foreground orange
	    } else {
		.f.t tag configure closeElement -foreground {}
	    }
	}

	empty {
	    if { $empty } {
		.f.t tag configure emptyElement -foreground green
	    } else {
		.f.t tag configure emptyElement -foreground {}
	    }
	}

	attributes {
	    if { $attributes } {
		.f.t tag configure attribute -foreground yellow
	    } else {
		.f.t tag configure attribute -foreground {}
	    }
	}

	hrefs {
	    if { $hrefs } {
		.f.t tag configure hrefs -underline true -foreground blue
		.f.t tag bind hrefs <Enter> \
		    [list .f.t configure -cursor center_ptr]
		.f.t tag bind hrefs <Leave> \
		    [list .f.t configure -cursor xterm]
		.f.t tag bind hrefs <Button-1> \
		    [list clickLink %x %y]
	    } else {
		.f.t tag configure hrefs -underline false -foreground {}
		.f.t tag bind hrefs <Enter> {}
		.f.t tag bind hrefs <Leave> {}
		.f.t tag bind hrefs <Button-1> {}
	    }
	}

    }

}


# This implements the "hyper link". It's called when someone clicks on a
# piece of text tagged as a link. It could do pretty much anything; this
# version just hands the link text to the mozilla browser.
#
# Parameters are the mouse x,y coordinates.
#
proc clickLink { xpos ypos } {

    # Find the text widget index of the character clicked on
    #
    set i [.f.t index @$xpos,$ypos]

    # Now find the range of characters spanned by the tag that char is in
    #
    set range [.f.t tag prevrange hrefs $i]

    # Fetching the string of characters within that range gives the URL
    #
    set url [eval .f.t get $range]

    # Execute a program interested in the URL
    #
    if { [catch { exec mozilla $url & }] } {
	.sbf.sb config -text "Unable to run mozilla!" -fg red
	after 3000 { .sbf.sb config -text "Ready" -fg black }
    }
}

# This is the procedure which scans the text in the text widget and sets the
# tags in place so the right parts of the text show up in the right colour.
# Ideally this would be run on each keypress so the text is always displayed
# with the right syntax highlights. On fast hardware that's probably possible,
# but on my PIII-500 machine it takes about half a second to run this on a 30K
# XML file. That's too slow for each keypress, so I use a timer to note when
# the user stops typing. The meaty bit of the code is only then run.
#
# Although it looks a bit heavy going, the code isn't actually that complex:
#
# Loop over all the text, starting at the top
#  Use a regex to find the next XML element - open, close or empty
#   Check to see if it's a close element - set the tag if so - done
#   Check to see if it's an empty element - set the tag if so - then continue
#   Loop over the element's attributes
#    Use a regular expression to pick out the name="value" string
#    Set the tag to indicate it's an attribute
#    Check to see if the attribute value starts with "http"
#     Set the tag to indicate it's an href attribute if it does
#  Move to the end of the processed element, then loop to look at the next one
#
set delayId ""
set currentText ""
proc applyHighlights { {delay 0} {force ""} } {
    global delayId currentText
    global open close empty attributes hrefs

    # Optimisation 1:
    #
    # The highlights are only applied after a small delay in the user's
    # typing: When this routine is called - i.e. on each keypress - a timer
    # is reset. If that timer is allowed to expire it means the user has
    # stopped typing for a moment so the highlights are recalculated. 

    # Cancel a running timer
    #
    if { $delayId != "" } {
	after cancel $delayId
    }
    
    # If a delay is required, restart the timer. The command to execute when
    # the timer expires is this routine, but without a delay.
    #
    if { $delay != 0 } {
	set delayId [after $delay [list applyHighlights 0]]
	return
    }
    set delayId ""

    # Optimisation 2:
    #
    # I store the complete text in a static string and compare that string
    # with the current text. If there hasn't been a change, nothing is done.
    # This means arrow keys don't cause the highlights to be reapplied.
    #
    # Tcl's "string equal" command is plenty fast enough to do this
    # comparison in just a few milliseconds - 4ms on my PIII-500 machine for
    # a 30K XML file.
    #
    if { $force != "true" &&
         [string equal $::currentText [.f.t get 1.0 end]] } {
	return
    }
    set currentText [.f.t get 1.0 end]


    # A 3rd optimisation would be to only apply the highlights in the
    # area the change has been made. That is non trivial, so I've not
    # implemented it for this example.


    # OK, we have to do it...
    #
    .sbf.sb configure -foreground red -text "Applying highlights"; update

    # Remove all the tags currently applied to the text
    #
    foreach tag [.f.t tag names] { .f.t tag remove $tag 1.0 end }

    set start "1.0"
    while { $start != "" } {

	# Use a regex to search for an open, close or empty XML tag. The count
	# option makes the command fill in a variable - c - with the number
	# of characters found which match the regex.
	#
	set start [.f.t search -count c -regexp --\
		                {<\/?[^>]+?\/?>} $start end]

	if { $start != "" } {
	    # Pull out the found text. It's faster to work on a string than
	    # keep referring back into the text widget
	    #
	    set foundText [.f.t get $start "$start +$c chars"]

	    # If the 2nd char found is "/" it's a close element tag: "</..."
	    #
	    if { [string equal [string index $foundText 1] "/"] } {
		.f.t tag add closeElement $start "$start +$c chars"
	    } else {

		# If the 2nd last char found is "/" it's an empty element,
		# otherwise it's a normal open element tag
		#
		if { [string equal [string index $foundText [expr $c-2]] \
                                   "/"] } {
		    .f.t tag add emptyElement $start "$start +$c chars"
		} else {
		    .f.t tag add openElement $start "$start +$c chars"
		}
	    
		# Check for attributes in the open element
		#
		set attStart $start
		set attEnd   [.f.t index "$attStart +$c chars"]

		# Loop over the element looking for name="value" sets
		#
		while { $attStart != "" } {
		    set attStart [.f.t search -count attC -regexp -- \
				        {\w+\s*="?[^"]+"?} $attStart $attEnd]
  		    if { $attStart != "" } {
                        # Set the tag to indicate this is an attribute
                        #
	                .f.t tag add attribute $attStart \
                                               "$attStart +$attC chars"

                        # Pull out the found attribute text so I can work on it
                        #
                        set found [.f.t get $attStart \
                                            "$attStart +$attC chars"]

                        if { [string match {href="http://*} $found] } {
			    set httpEnd [expr [string length $found] - 1]
			    .f.t tag add hrefs "$attStart +6 chars"\
			                        "$attStart +$httpEnd chars"
		        }

         	        set attStart [.f.t index "$attStart +$attC chars"]
	            }
	        }
	    }
	    set start [.f.t index "$start +$c chars"]
        }
    }

    .sbf.sb configure -foreground black -text "Ready"; update

}




#
# This is the start of the program...
#

# Check the parameter
#
if { [llength $argv] != 1 } {
    puts stderr "Usage: xmlviewer.tcl filename"
    exit 1
}

# Create the GUI. It is normal Tcl/Tk practise to store the names of important
# widgets in variables, then pass those variables around to the procedures
# which need them. This keeps the procedures free from hardcoded widget names,
# which are, at the end of the day, just global values themselves.
#
# For this example I've chosen not to do this because it makes the procedure
# calls look a little cluttered and I'm trying to keep things readable. The
# main text widget, created here, is called .f.t, and that name appears
# throughout the code.
#
wm title . "[lindex $argv 0] - XMLViewer.tcl"
pack [frame .f] -side top -fill both -expand true
text .f.t -setgrid true -wrap none \
	                -width 80 -height 24 \
	                -yscrollcommand ".f.sy set"
scrollbar .f.sy -orient v -command ".f.t yview"
pack .f.sy -side right -fill y
pack .f.t -side left -fill both -expand true

# .cf is the controls frame
#
pack [frame .cf] -side top -fill x -pady 3
pack [label .cf.l -text "Highlight:"] \
     [checkbutton .cf.open  -text "Open tags"  \
	  -variable open -command [list highlight "open"]] \
     [checkbutton .cf.close -text "Close tags" \
	  -variable close -command [list highlight "close"]] \
     [checkbutton .cf.empty -text "Empty tags" \
	  -variable empty -command [list highlight "empty"]] \
     [checkbutton .cf.attributes -text "Attributes" \
	  -variable attributes -command [list highlight "attributes"]] \
     [checkbutton .cf.hrefs -text "HTTP links" \
	  -variable hrefs -command [list highlight "hrefs"]] \
    -side left -ipady 3

# .sf is the search frame
#
pack [frame .sf] -side top -fill x -pady 3
pack [label .sf.l -text "Search (F3 to repeat): "] \
     [entry .sf.e -width 20] -side left -ipady 3

# .sbf is the status bar frame. The widget here is an "entry" widget, which
# is in many ways like a single line text widget. See the man page entry(n)
# for details.
#
pack [frame .sbf -relief sunken -bd 2] -side bottom -fill x
pack [label .sbf.sb -text "Ready"] -side left
bind .sf.e <KeyRelease> [list search %K]

focus .f.t

# Load the xml into the text widget
#
set status [catch {
    set handle [open [lindex $argv 0] "r"]
    .f.t insert end [read $handle]
    close $handle
} err]

if { $status } {
    .sbf.sb configure -text "Error reading data: $err"
}
tk::TextSetCursor .f.t 1.0

# Start with nothing highlighted
#
set open 0
set close 0
set empty 0
set attributes 0
set hrefs 0

# Create the tags for the text widget. Start with them all at default values
#
.f.t tag configure openElement
.f.t tag configure closeElement
.f.t tag configure emptyElement
.f.t tag configure attribute
.f.t tag configure hrefs
.f.t tag configure http
.f.t tag configure search -foreground white -background black

# First call to applyHightlights needs to happen with zero delay, and the force
# flag set to ensure the optimisations are bypassed.
#
applyHighlights 0 true

# applyHighlights will then be called on each keypress, but with a short
# delay before the scan actually takes place.
#
bind .f.t <KeyRelease> [list applyHighlights 1000]